/***************************************************************************
 *            SimTL.cc
 *
 *  Mon Apr 21 20:15:00 2008
 *  Copyright  2008  pchero21
 *  pchero21@gmail.com
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "SimTL.h"


void make_tmpdirec(void)
{
	char *dirPath = "/tmp/SimTL";	// 생성될 디렉토리 지정
	char *filename = "/tmp/SimTL/traceManager";	// 관리 파일 지정
	
	// 디렉토리를 생성한다. 실패하면 종료한다.
	if((mkdir(dirPath, 0755)) < 0) {
		system("/bin/rm -rf /tmp/SimTL");
		mkdir(dirPath, 0755);
	}

	// 디렉토리 속에 파일을 생성한다.
	p_option->fd_pckList = fopen(filename, "w");
	if(p_option->fd_pckList < 0) {
		printf("creat() failed!\n");
		exit(-1);
	}

	return;
}

FILE *make_tmpfile(FILE *fd_dir)
{
	char filename[50];
	struct timeval tv;
	time_t now;
	struct tm *tm_ptr;
	FILE *fd;
		
	time(&now);
	tm_ptr = localtime(&now);
	gettimeofday(&tv, NULL);
	
	bzero(filename, sizeof(filename));
	
	sprintf(filename, "/tmp/SimTL/%d.%d.%d_%d:%d:%d.%d",
		tm_ptr->tm_year + 1900,
		tm_ptr->tm_mon + 1,
		tm_ptr->tm_mday,
		tm_ptr->tm_hour,
		tm_ptr->tm_min,
		tm_ptr->tm_sec,
		tv.tv_usec);
	
	fd = fopen(filename, "wr");
	
	fprintf(stdout, "%d: ", p_option->chcnt++);
	filename[strlen(filename)] = '\t';
	fprintf(fd_dir, "%d: ", p_option->chcnt);
	fprintf(fd_dir, filename);
//	write(fd_dir, filename, strlen(filename));
	fprintf(stdout, filename, strlen(filename));
	
	return(fd);
}

void get_option(char **argv, pcap_option *p_option)
{
	int c;

	/* 옵션 초기화 */
	p_option->filter_rule = argv[1];		// ex) src host xxx.xxx.xxx.xxx and tcp port 80
	p_option->packetcnt = 20;				// 기본 캡처 20개
	p_option->pflag = 0;
	p_option->rflag = 0;
	p_option->eflag = 0;
	p_option->fflag = 0;
	p_option->cflag = 0;
	
	signal(SIGINT,sig_int);	// signal hanlder 등록 
	
	opterr = 0;
	
	while((c = getopt(p_option->argc, argv,"i:c:pfher")) != -1) {
		switch(c) {
			case 'i'  :			// 패킷 캡쳐 기기 지정 
				p_option->device = optarg;
				break;
			case 'c' : 		// 덤프하려는 패킷의 수
				p_option->cflag = 1; 
				p_option->packetcnt = atoi(optarg);
				if(p_option->packetcnt <= 0) {
					fprintf(stderr,"invalid pacet number %s",optarg);
					exit(1);
				} else if(p_option->packetcnt > 1000)
					p_option->packetcnt = 1000;
				break;
			case 'p' : 		// 데이터를 문자로 출력하는 옵션
				p_option->pflag = 1; 
				break;
			case 'f' :			// Ethernet Frame 출력
				p_option->fflag = 1;
				break;				
			case 'e' :      		// 데이터링크 계층 출력
				p_option->eflag = 1;
				break;		
			case 'r' :      		// 잡은 패킷을 맵핑없이 16진수로 모두 찍는다.
				p_option->rflag = 1;
				break;		
			case 'h' :			// 사용법
				usage();
				exit(1);
		}
	}
}

void usage(void)
{
    fprintf(stdout," Usage : SimTL_beta filter_rule [-pch]\n");
    fprintf(stdout,"         -p  :  데이타를 문자로 출력한다.\n");
    fprintf(stdout,"         -c  :  주어진 숫자만큼의 패킷만 덤프한다\n");
    fprintf(stdout,"	  -e  :  datalink layer를 출력한다.\n");
    fprintf(stdout,"	  -e  :  잡은 패킷을 생으로 찍는다.\n");
    fprintf(stdout,"         -h  :  사용법\n");
}

int packet_capture(pcap_option *p_option)
{
	struct bpf_program 	fcode;
	pcap_handler 				printer;
	char 							ebuf[PCAP_ERRBUF_SIZE];
	int 								i, snaplen = 512;
	bpf_u_int32 				localnet, netmask;
	unsigned char 			*pcap_userdata;
	static pcap_t				*pd;
	
	/* p_option 초기화 */
	p_option->chcnt = 0;
	
	make_tmpdirec();
	
	if (p_option->device == NULL ) {
	    if ( (p_option->device = pcap_lookupdev(ebuf) ) == NULL) {
	        perror(ebuf);           
   	        exit(-1);
   	    }
   	}
	fprintf(stdout, "device = %s\n", p_option->device);
   	
	
	pd = pcap_open_live(p_option->device, snaplen, PROMISCUOUS, 1000, ebuf);
	if(pd == NULL) {
		perror(ebuf);
		exit(-1);
	}
	
	i = pcap_snapshot(pd);
	if(snaplen < i) {
		perror("snaplen");
		perror(ebuf);
		exit(-1);
	}
	
	if(pcap_lookupnet(p_option->device, &localnet, &netmask, ebuf) < 0) {
		perror("pcap_lookupnet");
		perror(ebuf);
		exit(-1);
	}
	
	setuid(getuid());

	if(pcap_compile(pd, &fcode, p_option->filter_rule, 0, netmask) < 0) {
		perror(ebuf);
		exit(-1);
	}

	if(pcap_setfilter(pd, &fcode)< 0) {
		perror(ebuf);
		exit(-1);
	}
	
	fflush(stderr);
	
	printer = lookup_printer(pcap_datalink(pd));
	pcap_userdata = 0;

	if(pcap_loop(pd, p_option->packetcnt, printer, pcap_userdata) < 0) {
		perror("pcap_loop error");
		exit(-1);
	}
	
	pcap_close(pd);
	return(1);
}

static pcap_handler lookup_printer(int type)
{
	struct printer *p;
		
	for(p = printers; p->f; ++p)
		if(type == p->type)
			return p->f;
	perror("unknown data link type");
}

void packet_analysis(unsigned char *user, const struct pcap_pkthdr *h, const unsigned char *p)
{
	int									j, temp, chcnt;
	unsigned int					length = h->len;
	unsigned short			ether_type;
	unsigned char				*tcpdata, *udpdata, *icmpdata;
	register unsigned int	i;
	FILE								*fd_tmp;
	struct ether_header		*ep;
	struct iphdr					*iph;
	struct tcphdr					*tcph;
	struct udphdr					*udph;
	struct icmp					*icmph;
				
	chcnt = 0;
	fd_tmp = make_tmpfile(p_option->fd_pckList);
		
	// 잡은 패킷을 그대로 생으로 찍기	
	if(p_option->rflag) {
	    while(length--) {
	        fprintf(fd_tmp, "\t%X ", *(p++));
	        if( (++chcnt % 16) == 0 ) 
				fprintf(fd_tmp, "\n");
	    }
	    fprintf(fd_tmp, "\n");
	    return;
	}
	
	length -= sizeof(struct ether_header);
		
	// ethernet header mapping
	ep = (struct ether_header *)p;
	
	// ethernet header 14 bytes 를 건너 뛴 포인터
	p += sizeof(struct ether_header);
		
	// data link type
	ether_type = ntohs(ep->ether_type);
	
	// LAN frame 이 IEEE802 인 경우 ether_type 필드가 길이 필드가 된다.
	if(ether_type <= 1500) {
		;
	} else {
		if(p_option->eflag) {		// eflag 를 확인
			fprintf(fd_tmp, "\n\n======================= DataLink Layer =======================\n\t");
			for(j = 0; j < ETH_ALEN; j++) {
				fprintf(fd_tmp, "%X", ep->ether_dhost[j]);
				fprintf(p_option->fd_pckList, "%X", ep->ether_dhost[j]);
				fprintf(stdout, "%X", ep->ether_dhost[j]);
				if(j != 5) {
					fprintf(fd_tmp, " : ");
					fprintf(p_option->fd_pckList, " : ");
					fprintf(stdout, " : ");
				}
			}
			fprintf(fd_tmp, " -----> ");
			fprintf(p_option->fd_pckList, " -----> ");
			fprintf(stdout, " -----> ");
			for(j = 0; j < ETH_ALEN; j++) {
				fprintf(fd_tmp, "%X", ep->ether_shost[j]);
				fprintf(p_option->fd_pckList, "%X", ep->ether_shost[j]);
				fprintf(stdout, "%X", ep->ether_shost[j]);
				if(j != 5) {
					fprintf(fd_tmp, " : ");
					fprintf(p_option->fd_pckList, " : ");
					fprintf(stdout, " : ");
				}
			}
			fprintf(fd_tmp, "\n\tether_type -> %x\n", ntohs(ep->ether_type));
			fprintf(p_option->fd_pckList, "\n\tether_type -> %x\n", ntohs(ep->ether_type));
		//	fprintf(stdout, "\n\tether_type -> %x\n", ntohs(ep->ether_type));
		}
		
		iph = (struct iphdr *) p;
		i = 0;
		if(ntohs(ep->ether_type) == ETHERTYPE_IP) {		// IP 패킷인가?
			// packet capturing 한 것을 화면에 출력하는 부분
			fprintf(fd_tmp, "\n\n======================= IP HEADER =======================\n");
			fprintf(p_option->fd_pckList, " IP:: ");
			fprintf(stdout, " IP:: ");
			fprintf(fd_tmp, "\t%s -----> ", 						inet_ntoa(iph->saddr));
			fprintf(fd_tmp, "%s\n", 								inet_ntoa(iph->daddr));
			fprintf(fd_tmp, "\tVersion :	%d\n", 				iph->version);
			fprintf(fd_tmp, "\tHeader Length : 	%d\n", 	iph->ihl);
			fprintf(fd_tmp, "\tService : 	%#x\n", 			iph->tos);
			fprintf(fd_tmp, "\tTotal Length : 	%d\n", 		ntohs(iph->tot_len));
			fprintf(fd_tmp, "\tIdentification : %d\n", 		ntohs(iph->id));
			fprintf(fd_tmp, "\tFragment Offset : 	%d\n", 	ntohs(iph->frag_off));
			fprintf(fd_tmp, "\tTime to Live : 	%d\n", 		iph->ttl);
			fprintf(fd_tmp, "\tChecksum :	%d\n", 			ntohs(iph->check));
		
			/*요약 정보를 위한 부분 */
			fprintf(p_option->fd_pckList, "%s----->", 					inet_ntoa(iph->saddr));
			fprintf(p_option->fd_pckList, "%s ", 							inet_ntoa(iph->daddr));
			fprintf(p_option->fd_pckList, "Version:%d ", 				iph->version);
			fprintf(p_option->fd_pckList, "Header_Length:%d ", 	iph->ihl);
			fprintf(p_option->fd_pckList, "Service:%#x ", 			iph->tos);
			fprintf(p_option->fd_pckList, "Total_Length:%d\n", 	ntohs(iph->tot_len));
			
			fprintf(stdout, "%s----->", 				inet_ntoa(iph->saddr));
			fprintf(stdout, "%s ", 							inet_ntoa(iph->daddr));
			fprintf(stdout, "Version:%d ", 				iph->version);
			fprintf(stdout, "Header_Length:%d ", 	iph->ihl);
			fprintf(stdout, "Service:%#x ", 			iph->tos);
			fprintf(stdout, "Total_Length:%d\n", 	ntohs(iph->tot_len));

			
			
			/* packet 의 ip 부분을 건너 뛴 곳에서부터 tcp header의 시작이 된다. */
			if(iph->protocol == IPPROTO_TCP) {
				tcph = (struct tcphdr *) (p + iph->ihl * 4);
				// tcp data......
				tcpdata = (unsigned char *) (p + (iph->ihl * 4) + (tcph->doff * 4));
				fprintf(fd_tmp, "\n\n======================= TCP HEADER =======================\n");
				fprintf(fd_tmp, "\tSource Port : 	%d\n", ntohs(tcph->source));
				fprintf(fd_tmp, "\tDestination Port : 	%d\n", ntohs(tcph->dest));
				fprintf(fd_tmp, "\tSequence Number : 	%u\n", ntohl(tcph->seq));
				fprintf(fd_tmp, "\tAcknowledgement Number : 	%u\n", ntohl(tcph->ack_seq));
				fprintf(fd_tmp, "\tData Offset : 	%d\n", tcph->doff);
				fprintf(fd_tmp, "\tWindow : 	%d\n", ntohs(tcph->window));
				fprintf(fd_tmp, "\tURG : %d\tACK : %d\tPSH : %d\tRST : %d\tSYN : %d\tFIN : %d\n",
							tcph->urg, tcph->ack, tcph->psh, tcph->rst, tcph->syn, tcph->fin);
				fprintf(fd_tmp, "\n======================= TCP DATA(HEXA) =======================\n\t");
				chcnt = 0;
				for(temp = (iph->ihl * 4) + (tcph->doff * 4); temp <= ntohs(iph->tot_len) - 1; temp++) {
					fprintf(fd_tmp, "%02x ", *(tcpdata++));
					if((++chcnt % 16) == 0)
						fprintf(fd_tmp, "\n\t");
				}
			
				if(p_option->pflag) {		// pflag 를 확인
					fprintf(fd_tmp, "\n======================= TCP DATA(CHAR) =======================\n");
					tcpdata = (unsigned char *) ((p + iph->ihl * 4) + (tcph->doff * 4));
					for(temp = (iph->ihl * 4) + (tcph->doff * 4); temp <= ntohs(iph->tot_len) - 1; temp++)
						if(*(tcpdata) == 13 || *(tcpdata) == 0 || *(tcpdata) > 127) {
							fprintf(fd_tmp, ".");
							//tcpdata++;
							tcpdata++;
						} else if(*(tcpdata) == '\n') {
							fprintf(fd_tmp, "%c", *(tcpdata++));
						} else if(*(tcpdata) < 32) {
							fprintf(fd_tmp, ".");
							tcpdata++;
						}
						//else if((char)(*tcpdata) == NULL) {
							//fprintf(fd_tmp, "%c", '\0');
							//tcpdata++;
						//}
						else 
							fprintf(fd_tmp, "%c", *(tcpdata++));
				}
				fprintf(fd_tmp, "\n\t\t<<<<<<<<<< End of Data >>>>>>>>>>\n");
			} else if(iph->protocol == IPPROTO_UDP) {			// UDP 패킷인가?
				udph = (struct udphdr *) (p + iph->ihl * 4) + 8;
				fprintf(fd_tmp, "\n======================= UDP HEADER =======================\n");
				fprintf(fd_tmp, "\tSource Port : 	%d\n", ntohs(udph->source));
				fprintf(fd_tmp, "\tDestination Port : 	%d\n", ntohs(udph->dest));
				fprintf(fd_tmp, "\tLength : 	%d\n", ntohs(udph->len));
				fprintf(fd_tmp, "\tChecksum : 	%x\n", ntohs(udph->check));
				fprintf(fd_tmp, "\n======================= UDP DATA(HEXA) =======================\n\t");
				
				chcnt = 0;
				for(temp = (iph->ihl * 4) + 8; temp <= ntohs(iph->tot_len) - 1; temp++) {
					fprintf(fd_tmp, "%02x ", *(udpdata++));
					if((++chcnt % 16) == 0)
						fprintf(fd_tmp, "\n\t");
				}
			
				udpdata = (unsigned char *) (p + iph->ihl * 4) + 8;
				if(p_option->pflag) {		// pflag 를 확인
					fprintf(fd_tmp, "\n======================= UDP DATA(CHAR) =======================\n");
					for(temp = (iph->ihl * 4) + 8; temp <= ntohs(iph->tot_len) - 1; temp++)
						fprintf(fd_tmp, "%c", *(udpdata++));
				}
				fprintf(fd_tmp, "\n\t\t<<<<<<<<<< End of Data >>>>>>>>>>\n");
			} else if(iph->protocol == IPPROTO_ICMP) {		// ICMP 패킷인가?
				icmph = (struct icmp *) (p + iph->ihl * 4);
				icmpdata = (unsigned char *) (p + iph->ihl * 4) + 8;
				fprintf(fd_tmp, "\n\n======================= ICMP HEADER =======================\n");
				fprintf(fd_tmp, "\tType : 			%d\n", 		icmph->icmp_type);
				fprintf(fd_tmp, "\tCode : 			%d\n", 		icmph->icmp_code);
				fprintf(fd_tmp, "\tChecksum : 	%02x\n", 	icmph->icmp_cksum);
				fprintf(fd_tmp, "\tID : 				%d\n", 		icmph->icmp_id);
				fprintf(fd_tmp, "\tSeq : 			%d\n", 		icmph->icmp_seq);
				
				fprintf(fd_tmp, "\n======================= ICMP DATA(HEXA) =======================\n\t");
				chcnt = 0;
				for(temp = (iph->ihl * 4) + 8; temp <= ntohs(iph->tot_len) - 1; temp++) {
					fprintf(fd_tmp, "%02x ", *(icmpdata++));
					if((++chcnt % 16) == 0)
						fprintf(fd_tmp, "\n\t");
				}
				fprintf(fd_tmp, "\n\t\t<<<<<<<<<< End of Data >>>>>>>>>>\n");
			}
		}
	}

	fclose(fd_tmp);
}

void sig_int(int sig)
{
	tmp_save();

	printf("Bye!!\n");
	exit(0);
}

void tmp_save(void)
{
	char c, path[50], lpath[100];
	
	bzero(path, sizeof(path));
	
	printf("Do you want a Save? [Y/n] :");
	c = getc(stdin);
	
	if(c != 'n') {
		printf("Type the save path.\n");
		printf("Dir : ");
		scanf("%s", path);
		sprintf(lpath, "/bin/cp -f /tmp/SimTL/* %s", path);
		printf(lpath);
		system(lpath);
	}
	exit(0);
}
